exports.RepositoryCPS = RepositoryCPS

var GIT = require("nodegit")
var FS = require("fs")
var ERROR = require("./error.js")
var RV = require("./rendez_vous.js")
var UTIL = require("./util.js")

function RepositoryCPS (path, k, h) {
  GIT.Repo.open(path, ERROR.maybe(function (raw) {
    var o = {}
    o.path = path
    o.raw = raw
    o.get_master_commit_cps = function (k, h) {
      h = ERROR.chain(h, "get_head " + path)
      o.get_head_commits_cps(function (branches) {
        if (branches.master) {
          k(branches.master)
        } else if (branches.trunk) {
          k(branches.trunk)
        } else {
          o.get_current_commit_cps(k, h)
        }
      }, h)
    }
    o.get_current_commit_cps = function (k, h) {
      h = ERROR.chain(h, "get_current_commit_cps " + path)
      FS.readFile(path + "/.git/HEAD", {encoding: "utf-8"}, ERROR.maybe(function (ref_line) {
        var ref_path = path + "/.git/" + ref_line.replace("ref: ", "").replace("\n", "")
        FS.readFile(ref_path, {encoding: "utf-8"}, ERROR.maybe(function (sha) {
          raw.getCommit(sha, ERROR.maybe(function (raw_commit) {
            k(Commit(raw_commit, o))
          }, ERROR.chain(h, "cant get commit " + sha)))
        }, ERROR.chain(h, "cant read " + ref_path)))
      }, ERROR.chain(h, "cant read head files")))
    }
    o.get_head_commits_cps = function (k, h) {
      h = ERROR.chain(h, "get_head_commits_cps " + path)
      FS.readdir(path + "/.git/refs/heads", ERROR.maybe(function (files) {
        var hash = {}
        var rendez_vous = RV.RendezVous(files.length, function () { k(hash) }, ERROR.chain(h, "cant read head file"))
        files.forEach(function (file) {
          var head_path = path + "/.git/refs/heads/" + file
          FS.readFile(head_path, {encoding: "utf-8"}, ERROR.maybe(function (str) {
            raw.getCommit(str, ERROR.maybe(function (raw_commit) {
              hash[file] = Commit(raw_commit, o)
              rendez_vous.yes()
            }, ERROR.chain(rendez_vous.no, "cant get commit " + str)))
          }, ERROR.chain(rendez_vous.no, "cant read " + head_path)))
        })
      }, ERROR.chain(h, "cant read head directory")))
    }
    k(o)
  }, ERROR.chain(h, "RepositoryCPS " + path)))
}

// Use toString to compare oid!
// !!! Commit#messageEncoding() triggers a segmentation fault
// The following methods actually expects a callback (not mentioned in nodegit doc):
//   - Commit#getTree
//   - TreeEntry#getBlob

function Commit (raw, repo) {
  var oid = raw.oid()
  function get_tree_cps (k, h) {
    raw.getTree(ERROR.maybe(function (raw_tree) {
      k(Tree(raw_tree))
    }, ERROR.chain(h, "get_tree_cps " + oid + " " + repo.path)))
  }
  function get_parents_cps (k, h) {
    raw.getParents(ERROR.maybe(function (raws) {
      k(raws.map(function (raw) { return Commit(raw, repo) }))
    }, ERROR.chain("get_parent_cps " + oid + " " + repo.path)))
  }
  function get_parent_index_cps (parent, k, h) {
    h = ERROR.chain(h, "get_parent_index_cps " + repo.path + " " + oid + " " + parent.oid)
    var parent_count = raw.parentCount()
    function loop (i) {
      if (i < parent_count) {
        var oid = raw.parentId(i)
        if (oid == null) {
          h(new Error("error on getting parent oid for parent #" + i))
        } else if (parent.oid.toString() == oid.toString()) {
          k(i)
        } else {
          loop(i + 1)
        }
      } else {
        h(new Error("given commit is not a parent"))
      }
    }
    loop(0)
  }
  function get_patches_cps (parent, k, h) {
    h = ERROR.chain(h, "get_patches_cps " + repo.path + " " + oid)
    get_parent_index_cps(parent, function (index) {
      raw.getDiff(ERROR.maybe(function (diffs) {
        var patches = []
        for (var i = 0; i < diffs[index].size(); i++) {
          patches.push(Patch(diffs[index].patch(i).patch, repo))
        }
        k(patches)
      }, h))
    }, h)
  }
  var o = {}
  o.oid = oid
  if (raw.author()) { o.author_email = raw.author().email() }
  if (raw.committer()) { o.committer_email = raw.committer().email() }
  o.date = raw.date()
  o.message = raw.message()
  o.get_tree_cps = get_tree_cps
  o.get_parents_cps = get_parents_cps
  o.get_patches_cps = get_patches_cps
  o.log_cps = function (depth, k, h) {
    var h = ERROR.chain(h, "log_cps " + repo.path + " " + oid)
    console.log(">>>>>>>> COMMIT: " + date + " " + committer_email + " " + author_email + " <<<<<<<<<")
    console.log(message)
    get_tree_cps(function (tree) {
      tree.log_cps(depth + 1, function () {
        get_parents_cps(function (parents) {
          if (parents.length > 0) {
            get_patches_cps(parents[0], function (patches) {
              patches.forEach(function (patch) {
                console.log(">> PATCH: " + patch.status + " " + patch.old_path + " " + patch.new_path)
              }, h)
              parents[0].log_cps(depth, k, h)
            }, h)
          } else {
            k()
          }
        }, h)
      }, h)
    }, h)
  }
  return o
}

function Tree (raw) {
  var oid = raw.oid()
  var blob_entries = {}
  var tree_entries = {}
  for (var i = 0; i < raw.size(); i++) {
    entry = raw.entryByIndex(i)
    if (entry.isFile()) {
      blob_entries[entry.name()] = entry
    } else if (entry.isTree()) {
      tree_entries[entry.name()] = entry
    }
  }
  function compute_diffs_cps (tree, path, on_add, on_mod, on_del, k, h) {
    get_blobs_cps(function (my_blobs) {
      tree.get_blobs_cps(function (his_blobs) {
        for (name in my_blobs) {
          if (his_blobs[name]) {
            if (my_blobs[name].oid.toString() != his_blobs[name].oid.toString()) {
              on_mod(path+"/"+name, my_blobs[name], his_blobs[name])
            }
            delete his_blobs[name]
          } else {
            on_del(path+"/"+name, my_blobs[name])
          }
        }
        for (name in his_blobs) {
          on_add(path+"/"+name, his_blobs[name])
        }
        get_trees_cps(function (my_trees) {
          tree.get_trees_cps(function (his_trees) {
            var my_tree_names = Object.keys(my_trees)
            function loop_my(i) {
              if (i < my_tree_names.length) {
                var name = my_tree_names[i]
                if (his_trees[name]) {
                  function k_diff () {
                    delete his_trees[name]
                    setImmediate(loop_my, i+1)
                  }
                  my_trees[name].compute_diffs_cps(his_trees[name], path+"/"+name, on_add, on_mod, on_del, k_diff, h)
                } else {
                  my_trees[name].acc_blob_cps(path+"/"+name, function (path, blob, k_on, h) { on_del(path, blob); k_on() }, function () { setImmediate(loop_my, i+1) }, h)
                }
              } else {
                var his_tree_names = Object.keys(his_trees)
                function loop_his (i) {
                  if (i < his_tree_names.length) {
                    var name = his_tree_names[i]
                    his_trees[name].acc_blob_cps(path+"/"+name, function (path, blob, k_on, h) { on_add(path, blob); k_on() }, function () { setImmediate(loop_his, i+1) }, h) 
                  } else {
                    k()
                  }
                }
                loop_his(0)
              }
            }
            loop_my(0)
          }, h)
        }, h)
      }, h)
    }, h)
  }
  /*
  // concurrent version
          tree.get_trees_cps(function (his_trees) {
            var count = Object.keys(my_trees).length + Object.keys(his_trees).length
            var rv = RV.RendezVous(count, function () { k(adds, mods, dels) }, h)
            for (name in my_trees) {
              var child_path = path + "/" + name
              if (his_trees[name]) {
                my_trees[name].compute_diffs_cps(his_trees[name], child_path, function (child_adds, child_mods, child_dels) {
                  UTIL.merge_into(child_adds, adds)
                  UTIL.merge_into(child_mods, mods)
                  UTIL.merge_into(child_dels, dels)
                  rv.yes()
                }, rv.no)
                rv.yes()
                delete his_trees[name]
              } else {
                my_trees[name].acc_blob_cps(child_path, function (path, blob, k, h) {
                  dels[path] = blob
                  k()
                }, rv.yes, rv.no)
              }
            }
            for (name in his_trees) {
              his_trees[name].acc_blob_cps(path + "/" + name, function (path, blob, k, h) {
                  adds[path] = blob
                  k()
              }, rv.yes, rv.no)
            }
          }, h)
        }, h)
      }, h)
    }, h)
  }
  */
  function get_blobs_cps (k, h) {
    var hash = {}
    var rendez_vous = RV.RendezVous(Object.keys(blob_entries).length, function () { k(hash) }, ERROR.chain(h, "get_blobs_cps " + oid))
    for (var name in blob_entries) {
      each(name, blob_entries[name])
    }
    function each (name, entry) {
      entry.getBlob(ERROR.maybe(function (raw_blob) {
        hash[name] = Blob(raw_blob)
        rendez_vous.yes()
      }, ERROR.chain(rendez_vous.no, name)))
    }
  }
  function get_trees_cps (k, h) {
    var hash = {}
    var rendez_vous = RV.RendezVous(Object.keys(tree_entries).length, function () { k(hash) }, ERROR.chain(h, "get_trees_cps " + oid))
    for (var name in tree_entries) {
      each(name, tree_entries[name])
    }
    function each (name, entry) {
      entry.getTree(ERROR.maybe(function (raw_tree) {
        hash[name] = Tree(raw_tree)
        rendez_vous.yes()
      }, ERROR.chain(rendez_vous.no, name)))
    }
  }
  function acc_blob_cps (path, acc_cps, k, h) {
    h = ERROR.chain(h, "acc_blobs_cps " + path)
    get_blobs_cps(function (blobs) {
      get_trees_cps(function (trees) {
        var rendez_vous = RV.RendezVous(Object.keys(blobs).length + Object.keys(trees).length, k, h)
        UTIL.each(blobs, function (name, blob) {
          acc_cps(path + "/" + name, blobs[name], rendez_vous.yes, rendez_vous.no)
        })
        UTIL.each(trees, function (name, tree) {
          tree.acc_blob_cps(path + "/" + name, acc_cps, rendez_vous.yes, rendez_vous.no)
        })
      }, h)
    }, h)
  }
  function log_cps (depth, k, h) {
    get_blobs_cps(function (blobs) {
      for (name in blobs) {
        console.log(UTIL.padding(depth) + ">> BLOB: " + name + " <<")
        blobs[name].log(depth+1)
      }
      get_trees_cps(function (trees) {
        var names = Object.keys(trees)
        function loop(i) {
          if (i < names.length) {
            console.log(UTIL.padding(depth) + ">>>> TREE: " + names[i] + " <<<<")
            trees[names[i]].log_cps(depth+1, function () { loop(i+1) }, h)
          } else {
            k()
          }
        }
        loop(0)
      })
    })
  }
  var o = {}
  o.oid = oid
  o.get_blobs_cps = get_blobs_cps
  o.get_trees_cps = get_trees_cps
  o.acc_blob_cps = acc_blob_cps
  o.log_cps = log_cps
  o.compute_diffs_cps = compute_diffs_cps
  return o
}

function Blob (raw) {
  var o = {}
  o.oid = raw.oid()
  o.string = raw.toString()
  o.binary = raw.content()
  o.is_binary = raw.isBinary()
  o.file_mode = raw.filemode()
  o.log = function (depth) {
    var lines = raw.toString().split("\n")
    lines = lines.map(function(line) { return UTIL.padding(depth) + line })
    console.log(lines.join("\n"))
  }
  return o
}

/*

function patches (commit) {
  console.log("yo")
  commit.get_parents_cps(function (parents) {
    if (parents.length > 0) {
      commit.get_tree_cps(function (new_tree) {
        parents[0].get_tree_cps(function (old_tree) {
          console.log("swag")
          old_tree.compute_diffs_cps(new_tree, "", function (adds, mods, dels) {
            console.log(commit.date)
            console.log("ADDITIONS")
            console.log(adds)
            console.log("MODIFICATIONS")
            console.log(mods)
            console.log("DELETIONS")
            console.log(dels)
            patches(parents[0])
          }, ERROR.signal)
        }, ERROR.signal)
      }, ERROR.signal)
    }
  }, ERROR.signal)
}

var path = "/Users/soft/Desktop/test"
RepositoryCPS(path, function (repo) {
  repo.get_current_commit_cps(patches, ERROR.signal)
}, ERROR.signal)
*/

/*
//Test

var path = "/Users/soft/Desktop/test"
RepositoryCPS(path, function (repo) {
  repo.get_current_commit_cps(function (commit) {
    commit.log_cps(0, function () {
      commit.get_tree_cps(function (tree) {
        tree.acc_blob_cps("", function (path, blob, k, h) {
          console.log(path)
          console.log(blob.string)
          k()
        }, function () { console.log("END") }, ERROR.signal)
      }, ERROR.signal)
    }, ERROR.signal)
  }, ERROR.signal)
}, ERROR.signal)
*/